import typing

from dbnd._core.tracking.tracking_info_convertor import build_task_run_info
from dbnd._core.utils.json_utils import convert_to_safe_types
from dbnd.api.serialization.task import TaskRunInfoSchema
from dbnd_airflow_contrib.dbnd_operator import DbndOperator


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun


def build_dbnd_operator_from_taskrun(task_run):
    # type: (TaskRun)-> DbndOperator

    task = task_run.task
    params = convert_to_safe_types(
        {p.name: value for p, value in task._params.get_params_with_value()}
    )
    op_kwargs = task.task_airflow_op_kwargs or {}

    op = DbndOperator(
        task_id=task_run.task_af_id,
        dbnd_task_type=task.get_task_family(),
        dbnd_task_id=task.task_id,
        params=params,
        **op_kwargs
    )

    if task.task_retries is not None:
        op.retries = task.task_retries
        op.retry_delay = task.task_retry_delay

    task.ctrl.airflow_op = op
    set_af_operator_doc_md(task_run, op)
    return op


_task_run_info_schema = TaskRunInfoSchema()


def set_af_operator_doc_md(task_run, airflow_op):
    op = airflow_op
    op.doc_md = (
        "### Databand Info\n"
        "* **Tracker**: [{0}]({0})\n"
        "* **TaskRun UID**: {1}\n"
        "* **Run Name**: {2}\n"
        "* **Run UID**: {3}\n".format(
            task_run.task_tracker_url,
            task_run.task_run_uid,
            task_run.run.name,
            task_run.run.run_uid,
        )
    )
    op.dbnd_task_run_uid = task_run.task_run_uid
    op.dbnd_run_uid = task_run.run.run_uid
    op.dbnd_task_tracker_url = task_run.task_tracker_url
    op.dbnd_task_run_info = _task_run_info_schema.dump(
        build_task_run_info(task_run)
    ).data



import logging


def create_app(config=None, testing=False):
    from airflow.www_rbac import app as airflow_app

    app, appbuilder = airflow_app.create_app(config=config, testing=testing)

    # only now we can load view..
    # this import might causes circular dependency if placed above
    from dbnd_airflow.airflow_override.dbnd_aiflow_webserver import (
        use_databand_airflow_dagbag,
    )

    use_databand_airflow_dagbag()
    logging.info("Airflow applications has been created")
    return app, appbuilder


def cached_appbuilder(config=None, testing=False):
    _, appbuilder = create_app(config, testing)
    return appbuilder



from airflow import models
from airflow.configuration import conf as airflow_conf
from airflow.jobs.base_job import BaseJob
from airflow.utils import helpers
from airflow.utils.db import provide_session
from airflow.utils.state import State
from sqlalchemy import and_, or_

from dbnd_airflow.constants import AIRFLOW_VERSION_2


if AIRFLOW_VERSION_2:
    from airflow.utils.log.logging_mixin import LoggingMixin
    from airflow.models.taskinstance import SimpleTaskInstance
else:
    from airflow import LoggingMixin
    from airflow.utils.dag_processing import SimpleTaskInstance


def get_airflow_conf_log_folder() -> str:
    if AIRFLOW_VERSION_2:
        return airflow_conf.get("logging", "BASE_LOG_FOLDER")

    return airflow_conf.get("core", "BASE_LOG_FOLDER")


def get_airflow_conf_remote_logging() -> bool:
    if AIRFLOW_VERSION_2:
        return airflow_conf.getboolean("logging", "remote_logging")
    else:
        return airflow_conf.getboolean("core", "remote_logging")


def is_task_instance_finished(current_state: str) -> bool:
    # in AF 1.x this is a function returning array,
    #   while in 2.x - class variable
    if AIRFLOW_VERSION_2:
        return current_state in State.finished
    else:
        return current_state in State.finished()


if AIRFLOW_VERSION_2:
    from airflow.executors.local_executor import LocalExecutor
    from airflow.executors.sequential_executor import SequentialExecutor
else:
    from airflow.executors import LocalExecutor, SequentialExecutor


@provide_session
def reset_state_for_orphaned_tasks(
    single_dag_run_job: BaseJob, filter_by_dag_run=None, session=None
):
    """
    It was removed from Airflow 2.x Worth looking why it happened in Airflow github
    For now just replicated behaviour of v1. Probably not needed anymore

    This function checks if there are any tasks in the dagrun (or all)
    that have a scheduled state but are not known by the
    executor. If it finds those it will reset the state to None
    so they will get picked up again.
    The batch option is for performance reasons as the queries are made in
    sequence.

    :param filter_by_dag_run: the dag_run we want to process, None if all
    :type filter_by_dag_run: airflow.models.DagRun
    :return: the TIs reset (in expired SQLAlchemy state)
    :rtype: list[airflow.models.TaskInstance]
    """
    from airflow.jobs.backfill_job import BackfillJob

    queued_tis = single_dag_run_job.executor.queued_tasks
    # also consider running as the state might not have changed in the db yet
    running_tis = single_dag_run_job.executor.running

    resettable_states = [State.SCHEDULED, State.QUEUED]
    TI = models.TaskInstance
    DR = models.DagRun
    if filter_by_dag_run is None:
        resettable_tis = (
            session.query(TI)
            .join(
                DR,
                and_(TI.dag_id == DR.dag_id, TI.execution_date == DR.execution_date),
            )
            .filter(
                DR.state == State.RUNNING,
                DR.run_id.notlike(BackfillJob.ID_PREFIX + "%"),
                TI.state.in_(resettable_states),
            )
        ).all()
    else:
        resettable_tis = filter_by_dag_run.get_task_instances(
            state=resettable_states, session=session
        )
    tis_to_reset = []
    # Can't use an update here since it doesn't support joins
    for ti in resettable_tis:
        if ti.key not in queued_tis and ti.key not in running_tis:
            tis_to_reset.append(ti)

    if len(tis_to_reset) == 0:
        return []

    def query(result, items):
        filter_for_tis = [
            and_(
                TI.dag_id == ti.dag_id,
                TI.task_id == ti.task_id,
                TI.execution_date == ti.execution_date,
            )
            for ti in items
        ]
        reset_tis = (
            session.query(TI)
            .filter(or_(*filter_for_tis), TI.state.in_(resettable_states))
            .with_for_update()
            .all()
        )
        for ti in reset_tis:
            ti.state = State.NONE
            session.merge(ti)
        return result + reset_tis

    reset_tis = helpers.reduce_in_chunks(
        query, tis_to_reset, [], single_dag_run_job.max_tis_per_query
    )

    task_instance_str = "\n\t".join([repr(x) for x in reset_tis])
    session.commit()

    single_dag_run_job.log.info(
        "Reset the following %s TaskInstances:\n\t%s",
        len(reset_tis),
        task_instance_str,
    )
    return reset_tis



# This code was taken from SQLAlchemy-Utils version 0.36.1
# License is BSD

import os

from copy import copy

import sqlalchemy as sa

from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import OperationalError, ProgrammingError


def database_exists(url):
    """Check if a database exists.

    :param url: A SQLAlchemy engine URL.

    Performs backend-specific testing to quickly determine if a database
    exists on the server. ::

        database_exists('postgresql://postgres@localhost/name')  #=> False
        create_database('postgresql://postgres@localhost/name')
        database_exists('postgresql://postgres@localhost/name')  #=> True

    Supports checking against a constructed URL as well. ::

        engine = create_engine('postgresql://postgres@localhost/name')
        database_exists(engine.url)  #=> False
        create_database(engine.url)
        database_exists(engine.url)  #=> True

    """

    def get_scalar_result(engine, sql):
        result_proxy = engine.execute(sql)
        result = result_proxy.scalar()
        result_proxy.close()
        engine.dispose()
        return result

    def sqlite_file_exists(database):
        if not os.path.isfile(database) or os.path.getsize(database) < 100:
            return False

        with open(database, "rb") as f:
            header = f.read(100)

        return header[:16] == b"SQLite format 3\x00"

    url = copy(make_url(url))
    database = url.database
    if url.drivername.startswith("postgres"):
        url.database = "postgres"
    else:
        url.database = None

    engine = sa.create_engine(url)

    if engine.dialect.name == "postgresql":
        text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
        return bool(get_scalar_result(engine, text))

    elif engine.dialect.name == "mysql":
        text = (
            "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
            "WHERE SCHEMA_NAME = '%s'" % database
        )
        return bool(get_scalar_result(engine, text))

    elif engine.dialect.name == "sqlite":
        if database:
            return database == ":memory:" or sqlite_file_exists(database)
        else:
            # The default SQLAlchemy database is in memory,
            # and :memory is not required, thus we should support that use-case
            return True

    else:
        engine.dispose()
        engine = None
        text = "SELECT 1"
        try:
            url.database = database
            engine = sa.create_engine(url)
            result = engine.execute(text)
            result.close()
            return True

        except (ProgrammingError, OperationalError):
            return False
        finally:
            if engine is not None:
                engine.dispose()
